/*
Copyright (c) 2010-2011 The Regents of the University of California. All rights
reserved.

Permission is hereby granted, without written agreement and without license or 
royalty fees, to use, copy, modify, and distribute this software and its 
documentation for any purpose, provided that the above copyright notice and the
following two paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF 
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, 
AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/

/**
 * @file ptp-msg.c
 * Handle sending and receiving of PTP messages.
 *
 * @author Michael Zimmer (mzimmer@eecs.berkeley.edu)
 * @version $Id$
 */
 
#include "../ptp-protocol.h"
#include "DP83640.h"
#include "uip.h"
 
extern struct uip_udp_conn * ptp_event_conn; /**< uIP connection to send event 
* messages. */
extern struct uip_udp_conn * ptp_general_conn; /**< uIP connection to send
* general messages. */

// These macro's are implementation specific (depending on endianness of 
// processor)
// PTP packets are big-endian.
#define set08(x,y) (*(x) = (y)) /**< Set the 8-bits at 
* address \a x in big-endian to the UInteger8 or Integer8 variable \a y. 
*/
#define set16(x,y) (*(x) = ((y) >> 8), *((x) + 1) = (y)) /**< Set the 16-bits at
* address \a x in big-endian to the UInteger16 or Integer16 variable \a y. 
*/
#define set32(x,y) (*(x) = ((y) >> 24), *((x) + 1) = ((y) >> 16), \
    *((x) + 2) = ((y) >> 8), *((x) + 3) = (y)) /**< Set the 32-bits 
* at address \a x in big-endian to the UInteger32 or Integer32 variable \a y. 
*/ 
#define set64(x,y) (memcpy((x), &(y), 8)) /**< Set the 64-bits at address 
* \a x in big-endian to the UInteger64 or Integer64 variable \a y. 
*/
#define get4h(x,y) ((x) = (*(y) >> 4) & 0x0F) /**< Set the variable \a x to
* the upper 4-bits in the 8-bit value in big-endian at address \a y.
*/
#define get4l(x,y) ((x) = (*(y)) & 0x0F) /**< Set the variable \a x to
* the lower 4-bits in the 8-bit value in big-endian at address \a y.
*/
#define get08(x,y) ((x) = *(y)) /**< Set the UInteger8 or Integer8 variable
* \a x to the upper 4-bits in the 8-bit value in big-endian at address \a y.
*/
#define get16(x,y) (memcpy(&(x), (y), 2))  /**< Set the UInteger16 or 
* Integer16 variable \a x to the 16-bit value in big-endian at address \a y.
*/
#define get32(x,y) (memcpy(&(x), (y), 4))  /**< Set the UInteger32 or 
* Integer32 variable \a x to the 32-bit value in big-endian at address \a y.
*/
#define get64(x,y) (memcpy(&(x), (y), 8)) /**< Set the UInteger64 or Integer64
* variable \a x to the 64-bit value in big-endian at address \a y.
*/

/** Send Announce message over the network as a multicast PTP general message. 
* This implementation uses current data from the PTPState \a ptp and the 
* current time to fill the buffer location \a uip_appdata with the Announce 
* message. Once the message is correctly in the buffer, the uIP protocol stack 
* is called to send the message.
*/
void sendAnnounceMsg(PTPState *ptp)
{
    
    UInteger8 *buf;
    Timestamp timestamp;

    buf = (UInteger8 *)uip_appdata;
    getClock(&timestamp);
    
    // Increment sequence Id.
    ptp->sequenceIdAnnounce++;
    
    // Message header (13.3.1).
    set08(buf + 0, (TRANSPORTSPECIFIC << 4) | ANNOUNCE);
    set08(buf + 1, VERSIONPTP);
    set16(buf + 2, MESSAGELENGTH_ANNOUNCE);
    set08(buf + 4, ptp->defaultDS.domainNumber);
    set08(buf + 6, (UNICASTFLAG << 2) | ALTERNATEMASTERFLAG);
    set08(buf + 7, ptp->timePropertiesDS.leap61 |
                  (ptp->timePropertiesDS.leap59 << 1) |
                  (ptp->timePropertiesDS.currentUtcOffsetValid << 2) |
                  (ptp->timePropertiesDS.ptpTimescale << 3) |
                  (ptp->timePropertiesDS.timeTraceable << 4) |
                  (ptp->timePropertiesDS.frequencyTraceable << 5));
    memset(buf + 8, 0, 8);
    memcpy(buf + 20, ptp->portDS.portIdentity.clockIdentity, 8);
    
    set16(buf + 28, ptp->portDS.portIdentity.portNumber);
    set16(buf + 30, ptp->sequenceIdAnnounce);
    set08(buf + 32, CONTROLFIELD_ALL_OTHERS);
    set08(buf + 33, ptp->portDS.logAnnounceInterval);

    // Announce message fields (13.5).
    set16(buf + 34, timestamp.secondsField.msb);
    set32(buf + 36, timestamp.secondsField.lsb);
    set32(buf + 40, timestamp.nanosecondsField);
    set16(buf + 44, ptp->timePropertiesDS.currentUtcOffset);
    set08(buf + 47, ptp->parentDS.priority1);
    set08(buf + 48, ptp->parentDS.grandmasterClockQuality.clockClass);
    set08(buf + 49, ptp->parentDS.grandmasterClockQuality.clockAccuracy);
    set16(buf + 50, 
          ptp->parentDS.grandmasterClockQuality.offsetScaledLogVariance);
    set08(buf + 52, ptp->parentDS.priority2);
    memcpy(buf + 53, ptp->parentDS.grandmasterIdentity, 8);
    set16(buf + 61, ptp->currentDS.stepsRemoved);
    set08(buf + 63, ptp->timePropertiesDS.timeSource);

    // Send Announce message.
    uip_udp_conn = ptp_general_conn;
    uip_udp_send(MESSAGELENGTH_ANNOUNCE);
    
}

/** Send Sync message over the network as a multicast PTP event message. 
* This implementation uses current data from the PTPState \a ptp and the 
* current time to fill the buffer location \a uip_appdata with the Sync 
* message. Once the message is correctly in the buffer, the uIP protocol
* stack is called to send the message.
*/
void sendSyncMsg(PTPState *ptp)
{
    
    UInteger8 *buf;
    Timestamp timestamp;

    buf = (UInteger8 *)uip_appdata;

    // Increment sequence Id.
    ptp->sequenceIdSync++;

    // Message header (13.3.1).
    set08(buf + 0, (TRANSPORTSPECIFIC << 4) | SYNC);
    set08(buf + 1, VERSIONPTP);
    set16(buf + 2, MESSAGELENGTH_SYNC);
    set08(buf + 4, ptp->defaultDS.domainNumber);
    set08(buf + 6, (UNICASTFLAG << 2) | (TWOSTEPFLAG << 1) | 
                   ALTERNATEMASTERFLAG);
    memset(buf + 8, 0, 8); 
    memcpy(buf + 20, ptp->portDS.portIdentity.clockIdentity, 8);
    set16(buf + 28, ptp->portDS.portIdentity.portNumber);
    set16(buf + 30, ptp->sequenceIdSync);
    set08(buf + 32, CONTROLFIELD_SYNC);
    set08(buf + 33, ptp->portDS.logSyncInterval);

    // Sync message fields. (13.6)
    if(ptp->defaultDS.twoStepFlag == TRUE) {
        getClock(&timestamp);    
        set16(buf + 34, timestamp.secondsField.msb);
        set32(buf + 36, timestamp.secondsField.lsb);
        set32(buf + 40, timestamp.nanosecondsField);  
    }  
    else {
        memset((buf + 34), 0, 10); // DP83640 will fill this packet.   
    }

    // Send Sync message.
    uip_udp_conn = ptp_event_conn;
    uip_udp_send(MESSAGELENGTH_SYNC);

}

/** Send Follow_Up message over the network as a multicast PTP general message.
* This implementation uses current data from the PTPState \a ptp to fill the 
* buffer location \a uip_appdata with the Follow_Up message. Once the message 
* is correctly in the buffer, the uIP protocol stack is called to send the 
* message.
*/
void sendFollowUpMsg(PTPState *ptp)
{

    UInteger8 *buf;
    buf = (UInteger8 *)uip_appdata;

    // Message header (13.3.1).
    set08(buf + 0, (TRANSPORTSPECIFIC << 4) | FOLLOW_UP);
    set08(buf + 1, VERSIONPTP);
    set16(buf + 2, MESSAGELENGTH_FOLLOW_UP);
    set08(buf + 4, ptp->defaultDS.domainNumber);
    set08(buf + 6, (UNICASTFLAG << 2) | ALTERNATEMASTERFLAG);
    memset(buf + 8, 0, 8); 
    memcpy(buf + 20, ptp->portDS.portIdentity.clockIdentity, 8);
    set16(buf + 28, ptp->portDS.portIdentity.portNumber);
    set16(buf + 30, ptp->sequenceIdSync);
    set08(buf + 32, CONTROLFIELD_FOLLOW_UP);
    set08(buf + 33, 0);

    // Follow_Up message fields. (13.7)
    set16(buf + 34, ptp->syncEventEgressTimestamp.secondsField.msb);
    set32(buf + 36, ptp->syncEventEgressTimestamp.secondsField.lsb);
    set32(buf + 40, ptp->syncEventEgressTimestamp.nanosecondsField);    

    // Send Follow_Up message.
    uip_udp_conn = ptp_general_conn;
    uip_udp_send(MESSAGELENGTH_FOLLOW_UP);
    
}

/** Send Delay_Req message over the network as a multicast PTP event message.
* This implementation uses current data from the PTPState \a ptp to fill the 
* buffer location \a uip_appdata with the Delay_Req message. Once the message 
* is correctly in the buffer, the uIP protocol stack is called to send the 
* message.
*/
void sendDelayReqMsg(PTPState *ptp)
{

    UInteger8 *buf;
    Integer64 c = -DELAY_ASYMMETRY;

    buf = (UInteger8 *)uip_appdata;

    // Increment sequence Id.
    ptp->sequenceIdDelayReq++;

    // Message header (13.3.1).
    set08(buf + 0, (TRANSPORTSPECIFIC << 4) | DELAY_REQ);
    set08(buf + 1, VERSIONPTP);
    set16(buf + 2, MESSAGELENGTH_DELAY_REQ);
    set08(buf + 4, ptp->defaultDS.domainNumber);
    set08(buf + 6, (UNICASTFLAG << 2));
    set64(buf + 8, c);
    memcpy(buf + 20, ptp->portDS.portIdentity.clockIdentity, 8);
    set16(buf + 28, ptp->portDS.portIdentity.portNumber);
    set16(buf + 30, ptp->sequenceIdDelayReq);
    set08(buf + 32, CONTROLFIELD_DELAY_REQ);
    set08(buf + 33, 0x7F);
    
    // Delay_Req message fields. (13.6)
    memset((buf + 34), 0, 10); 
    
    // Send Delay_Req message.
    uip_udp_conn = ptp_event_conn;
    uip_udp_send(MESSAGELENGTH_DELAY_REQ);

}

/** Send Delay_Resp message over the network as a multicast PTP general message.
* This implementation
* uses current data from the PTPState \a ptp and the DelayReqMsg \a msg for
* which this message is a response to fill the buffer location
* \a uip_appdata with the Delay_Resp message. Once the message is
* correctly in the buffer, the uIP protocol stack is called to send the message.
*/
void sendDelayRespMsg(PTPState *ptp, DelayReqMsg *msg)
{

    UInteger8 *buf;

    buf = (UInteger8 *)uip_appdata;

    // Message header (13.3.1).
    set08(buf + 0, (TRANSPORTSPECIFIC << 4) | DELAY_RESP);
    set08(buf + 1, VERSIONPTP);
    set16(buf + 2, MESSAGELENGTH_DELAY_RESP);
    set08(buf + 4, ptp->defaultDS.domainNumber);
    set08(buf + 6, (UNICASTFLAG << 2) | ALTERNATEMASTERFLAG);
    set64(buf + 8, msg->header.correctionField);
    memcpy(buf + 20, ptp->portDS.portIdentity.clockIdentity, 8);
    set16(buf + 28, ptp->portDS.portIdentity.portNumber);
    set16(buf + 30, msg->header.sequenceId);
    set08(buf + 32, CONTROLFIELD_DELAY_RESP);
    set08(buf + 33, ptp->portDS.logMinDelayReqInterval);

    // Delay_Req message fields. (13.8)
    set16(buf + 34, ptp->delayEventIngressTimestamp.secondsField.msb);
    set32(buf + 36, ptp->delayEventIngressTimestamp.secondsField.lsb);
    set32(buf + 40, ptp->delayEventIngressTimestamp.nanosecondsField);
    memcpy(buf + 44, msg->header.sourcePortIdentity.clockIdentity, 8);    
    set16(buf + 52, msg->header.sourcePortIdentity.portNumber);

    // Send Delay_Resp message.
    uip_udp_conn = ptp_general_conn;
    uip_udp_send(MESSAGELENGTH_DELAY_RESP);
    
}

/** Send Management message being sent over the network as a multicast PTP 
* general message. This implementation uses current data from the PTPState 
* \a ptp and the ManagementMsg \a msg which is being sent to fill the buffer 
* location \a uip_appdata with the Management message. Once the message is 
* correctly in the buffer, the uIP protocol stack is called to send the 
* message.
*/
void sendManagementMsg(PTPState *ptp, ManagementMsg *msg)
{
    
    UInteger8 *buf;

    buf = (UInteger8 *)uip_appdata;

    // Message header (13.3.1).
    set08(buf + 0, (TRANSPORTSPECIFIC << 4) | MANAGEMENT);
    set08(buf + 1, VERSIONPTP);
    set16(buf + 2, (MESSAGELENGTH_MANAGEMENT + 
                   msg->managementTLV.dataFieldLength));
    set08(buf + 4, msg->header.domainNumber);
    set08(buf + 6, (UNICASTFLAG << 2));
    set64(buf + 8, msg->header.correctionField);
    memcpy(buf + 20, ptp->portDS.portIdentity.clockIdentity, 8);
    set16(buf + 28, ptp->portDS.portIdentity.portNumber);
    set16(buf + 30, msg->header.sequenceId);
    set08(buf + 32, CONTROLFIELD_MANAGEMENT);
    set08(buf + 33, 0x7F);    
    
    // Management message fields (15.4.1).
    memcpy(buf + 34, msg->targetPortIdentity.clockIdentity, 8);
    set16(buf + 42, msg->targetPortIdentity.portNumber);
    set08(buf + 44, msg->startingBoundaryHops);
    set08(buf + 45, msg->boundaryHops);
    set08(buf + 46, msg->actionField);
    
    // Management TLV fields (15.5.2).
    set16(buf + 48, msg->managementTLV.tlvType);
    set16(buf + 50, msg->managementTLV.lengthField);
    set16(buf + 52, msg->managementTLV.managementId);
    memcpy(buf + 54, msg->managementTLV.dataField, 
                     msg->managementTLV.dataFieldLength);
    
    // Send Management message.
    uip_udp_conn = ptp_general_conn;
    uip_udp_send(MESSAGELENGTH_MANAGEMENT + msg->managementTLV.dataFieldLength);
    
}


/** Attempt to provide an accurate timestamp for Sync message egress. */
void getSyncEventEgressTimestamp(PTPState *ptp)
{
    
    UInteger8 i;
    UInteger16 reg;
    
    // Use latest timestamp in hardware buffer.
    _phy_write(PAGE_SEL_REG, PTP_STS_PAGE);
    
    // Buffer can contain at 4 timestamps.
    for(i = 0; i < 4; i++) {
    
        // Check for timestamp in buffer.
        if((_phy_read(PTP_STS_REG) & 0x0800) != 0) {
        
            // Read timestamp (DP83640, 14.5.9).
            reg = _phy_read(PTP_TXTS_REG);
            ptp->syncEventEgressTimestamp.nanosecondsField = 
                    (reg & 0xFFFF) | (_phy_read(PTP_TXTS_REG) << 16);
            reg = _phy_read(PTP_TXTS_REG);
            ptp->syncEventEgressTimestamp.secondsField.lsb = 
                    (reg & 0xFFFF) | (_phy_read(PTP_TXTS_REG) << 16);
            ptp->syncEventEgressTimestamp.secondsField.msb = 0;
        
        }
     
    }    

}



/** Attempt to provide an accurate timestamp for Sync message ingress. */
Boolean getSyncEventIngressTimestamp(PTPState *ptp, UInteger16 seqId)
{

    UInteger16 reg;
    UInteger8 i;
    Timestamp timestamp;

    // Use latest timestamp in buffer.
    _phy_write(PAGE_SEL_REG, PTP_STS_PAGE);

    // Buffer could contain at most 4 timestamps.
    for(i = 0; i < 4; i++) {

        // Check for timestamp in buffer.
        if((_phy_read(PTP_STS_REG) & 0x0400) != 0) {

            // Read timestamp (DP83640, 14.5.10).
            reg = _phy_read(PTP_RXTS_REG);
            timestamp.nanosecondsField = 
                (0xFFFF & reg) | (_phy_read(PTP_RXTS_REG) << 16);
            reg = _phy_read(PTP_RXTS_REG);
            timestamp.secondsField.lsb = 
                (0xFFFF & reg) | (_phy_read(PTP_RXTS_REG) << 16);
            timestamp.secondsField.msb = 0;
            reg = _phy_read(PTP_RXTS_REG);
            if(reg == seqId && 
                (_phy_read(PTP_RXTS_REG) >> 12) == SYNC) {
                ptp->syncEventIngressTimestamp = timestamp;
                return TRUE;
            }

        }
        else {
            break;
        }
    }
    return FALSE;
    
}

/** Attempt to provide an accurate timestamp for Delay_Req message egress. */
void getDelayEventEgressTimestamp(PTPState *ptp)
{
    
    UInteger8 i;
    UInteger16 reg;
    
    // Use latest timestamp in hardware buffer.
    _phy_write(PAGE_SEL_REG, PTP_STS_PAGE);
    
    // Buffer can contain at 4 timestamps.
    for(i = 0; i < 4; i++) {
    
        // Check for timestamp in buffer.
        if((_phy_read(PTP_STS_REG) & 0x0800) != 0) {
        
            // Read timestamp (DP83640, 14.5.9).
            reg = _phy_read(PTP_TXTS_REG);
            ptp->delayEventEgressTimestamp.nanosecondsField = 
                    (reg & 0xFFFF) | (_phy_read(PTP_TXTS_REG) << 16);
            reg = _phy_read(PTP_TXTS_REG);
            ptp->delayEventEgressTimestamp.secondsField.lsb = 
                    (reg & 0xFFFF) | (_phy_read(PTP_TXTS_REG) << 16);
            ptp->delayEventEgressTimestamp.secondsField.msb = 0;
        
        }
     
    }
    
}

/** Attempt to provide an accurate timestamp for Delay_Req message ingress. */
Boolean getDelayEventIngressTimestamp(PTPState *ptp, UInteger16 seqId)
{
  
    UInteger8 i;
    UInteger16 reg;
    Timestamp timestamp;
    
    // Use latest timestamp in buffer.
    _phy_write(PAGE_SEL_REG, PTP_STS_PAGE);

    // Buffer could contain at most 4 timestamps.
    for(i = 0; i < 4; i++) { 

        // Check for timestamp in buffer.
        if((_phy_read(PTP_STS_REG) & 0x0400) != 0) {

            // Read timestamp (DP83640, 14.5.10).
            reg = _phy_read(PTP_RXTS_REG);
            timestamp.nanosecondsField =
                (0xFFFF & reg) | (_phy_read(PTP_RXTS_REG) << 16);
            reg = _phy_read(PTP_RXTS_REG);
            timestamp.secondsField.lsb = 
                (0xFFFF & reg) | (_phy_read(PTP_RXTS_REG) << 16);
            timestamp.secondsField.msb = 0;
            reg = _phy_read(PTP_RXTS_REG);
            if(reg == seqId && 
                (_phy_read(PTP_RXTS_REG) >> 12) == DELAY_REQ) {
                ptp->delayEventIngressTimestamp = timestamp;
                return TRUE;
            }
        }
        else {
            break;
        }
    }
    return FALSE;


}

/** Fill the provided Header data structure 
 * with data from the message. 
 */
void readHeader(Header *header) {

    UInteger8 *buf;

    buf = (UInteger8 *)uip_appdata;
    
    get4h(header->transportSpecific, buf + 0);
    get4l(header->messageType, buf + 0);
    get4l(header->versionPTP, buf + 1);
    get16(header->messageLength, buf + 2);
    get08(header->domainNumber, buf + 4);
    get08(header->flagField[0], buf + 6);
    get08(header->flagField[1], buf + 7);
    get64(header->correctionField, buf + 8);
    memcpy(header->sourcePortIdentity.clockIdentity, buf + 20, 8);
    get16(header->sourcePortIdentity.portNumber, buf + 28);
    get16(header->sequenceId, buf + 30);
    get08(header->controlField, buf + 32);
    get08(header->logMessageInterval, buf + 33);
    
}

/** Fill the provided AnnounceMsg data structure 
 * with data from an Announce message.
 */
void readAnnounceMsg(AnnounceMsg *msg) {
 
    UInteger8 *buf;

    buf = (UInteger8 *)uip_appdata;

    readHeader(&msg->header);
    get16(msg->originTimestamp.secondsField.msb, buf + 34);
    get32(msg->originTimestamp.secondsField.lsb, buf + 36);
    get32(msg->originTimestamp.nanosecondsField, buf + 40);
    get16(msg->currentUtcOffset, buf + 44);
    get08(msg->grandmasterPriority1, buf + 47);
    get08(msg->grandmasterClockQuality.clockClass, buf + 48);
    get08(msg->grandmasterClockQuality.clockAccuracy, buf + 49);
    get16(msg->grandmasterClockQuality.offsetScaledLogVariance, buf + 50);
    get08(msg->grandmasterPriority2, buf + 52);
    memcpy(msg->grandmasterIdentity, buf + 53, 8);
    get16(msg->stepsRemoved, buf + 61);
    get08(msg->timeSource, buf + 63);
    
}

/** Fill the provided SyncMsg data structure
 * with data from a Sync message.
 */
void readSyncMsg(SyncMsg *msg){
   
    UInteger8 *buf;

    buf = (UInteger8 *)uip_appdata;

    readHeader(&msg->header);
    get16(msg->originTimestamp.secondsField.msb, buf + 34);
    get32(msg->originTimestamp.secondsField.lsb, buf + 36);
    get32(msg->originTimestamp.nanosecondsField, buf + 40);
    
}

/** Fill the provided FollowUpMsg data structure 
 * with data from a Follow_Up message.
 */
void readFollowUpMsg(FollowUpMsg *msg) {

    UInteger8 *buf;

    buf = (UInteger8 *)uip_appdata;

    readHeader(&msg->header);
    get16(msg->preciseOriginTimestamp.secondsField.msb, buf + 34);
    get32(msg->preciseOriginTimestamp.secondsField.lsb, buf + 36);
    get32(msg->preciseOriginTimestamp.nanosecondsField, buf + 40);
    
}

/** Fill the provided DelayReqMsg data structure
 * with data from a Delay_Req message. 
 */
void readDelayReqMsg(DelayReqMsg *msg) {

    UInteger8 *buf;

    buf = (UInteger8 *)uip_appdata;

    readHeader(&msg->header);
    get16(msg->originTimestamp.secondsField.msb, buf + 34);
    get32(msg->originTimestamp.secondsField.lsb, buf + 36);
    get32(msg->originTimestamp.nanosecondsField, buf + 40);
    
}

/** Fill the provided DelayRespMsg data structure
 * with data from a Delay_Resp message.
 */
void readDelayRespMsg(DelayRespMsg *msg) {

    UInteger8 *buf;

    buf = (UInteger8 *)uip_appdata;

    readHeader(&msg->header);
    get16(msg->receiveTimestamp.secondsField.msb, buf + 34);
    get32(msg->receiveTimestamp.secondsField.lsb, buf + 36);
    get32(msg->receiveTimestamp.nanosecondsField, buf + 40);
    memcpy(msg->requestingPortIdentity.clockIdentity, buf + 44, 8);
    get16(msg->requestingPortIdentity.portNumber, buf + 52);
    
}

/** Fill the provided ManagementMsg data structure
 * with data from a Management message.
 */
void readManagementMsg(ManagementMsg *msg) {
    
    UInteger8 *buf;

    buf = (UInteger8 *)uip_appdata;

    readHeader(&msg->header);
    memcpy(msg->targetPortIdentity.clockIdentity, buf + 34, 8);
    get16(msg->targetPortIdentity.portNumber, buf + 42);
    get08(msg->startingBoundaryHops, buf + 44);
    get08(msg->boundaryHops, buf + 45);
    get4l(msg->actionField, buf + 46);
    get16(msg->managementTLV.tlvType, buf + 48);
    get16(msg->managementTLV.lengthField, buf + 50);
    get16(msg->managementTLV.managementId, buf + 52);
    msg->managementTLV.dataFieldLength = msg->header.messageLength - 
                                         MESSAGELENGTH_MANAGEMENT;
    memcpy(msg->managementTLV.dataField, buf + 54, 
           msg->managementTLV.dataFieldLength);
    
}

/** In SLAVE mode, the Delay_Req messages from other SLAVE clocks 
 * may be timestamped in hardware, even though they are ignored in software.
 * Nothing is done because it is assumed that messages are received in the
 * same order in software that they are timestamped in hardware, so 
 * ingressTimestamp is overwritten when the next message is processed.
 */
void clearIngressTimestamp() {
    
}


