/*
Copyright (c) 2010-2011 The Regents of the University of California. All rights
reserved.

Permission is hereby granted, without written agreement and without license or 
royalty fees, to use, copy, modify, and distribute this software and its 
documentation for any purpose, provided that the above copyright notice and the
following two paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF 
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, 
AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/

/**
 * @file ptp-data-types.h
 * Define all platform-independent data types and sets that are required for 
 * the correct operation of PTP. 
 *
 * @author Michael Zimmer (mzimmer@eecs.berkeley.edu)
 * @version $Id$
 */

/** A time interval (5.3.2). */
typedef struct 
{
    Integer64 scaledNanoseconds; /**< Time interval is expressed in units
    * of nanoseconds multiplied by 2^16 (5.3.2). If the time interval is 
    * outside the range of Integer64, INT64_MIN or INT64_MAX should be used 
    * instead.
    */
} TimeInterval;

/** A positive time with respect to the epoch (5.3.3). Epoch used depends 
 * on the timescale used by the grandmaster clock of the domain (Annex B).
 */
typedef struct
{
    UInteger48 secondsField; /**< Integer portion of the timestamp in units
    * of seconds (5.3.3).
    */
    UInteger32 nanosecondsField; /**< Fractional portion of the timestamp
    * in units of nanoseconds, always less than 10^9 (5.3.3).
    */
} Timestamp;

/** Identity of a clock (5.3.4). The value depends on implementation (7.5.2.2). 
 */
typedef Octet ClockIdentity[8];

/** Identity of a PTP port (5.3.5). */
typedef struct
{
    ClockIdentity clockIdentity; /**< Clock identity of the port (5.3.5,
    * 7.5.2.2).
    */
    UInteger16 portNumber; /**< Port number of the port (5.3.5, 7.5.2.3). 
    Ports numbers for a PTP node start at 1 and increment with number of ports.
    */
} PortIdentity;

/** Protocol address of a PTP port (5.3.6). */
typedef struct
{
    Enumeration16 networkProtocol; /**< Network protocol enumeration
    * (5.3.6, 7.4.1).
    */
    UInteger16 addressLength; /**< Length of the address in octets, in the
    * range 1 to 16 octets (5.3.6).
    */
    Octet* addressField; /**< Protocol address of a port (5.3.6). See 
    * applicable annex.
    */
} PortAddress;

/** Quality of a clock (5.3.7). */
typedef struct 
{
    UInteger8 clockClass; /**< Traceability of a clock (5.3.7, 7.6.2.4). */
    Enumeration8 clockAccuracy; /**< Accuracy of a clock (5.3.7, 7.6.2.5).
    */
    UInteger16 offsetScaledLogVariance; /**< Estimate of inherent precision of
    * a clock (5.3.7, 7.6.3.5).
    */
} ClockQuality;

/** A TLV extension (5.3.8). */
typedef struct
{
    Enumeration16 tlvType; /**< TLV type (5.3.8, 14.1.1). */
    UInteger16 lengthField; /**< Length of the TLV field in number of
    * of octets (must be even) (5.3.8, 14.1.2). */
    Octet* valueField; /**< Format and meaning of the value field is
    * defined by the TLV type (5.3.8, 14.1.3) .
    */
} TLV;

/** Text in a PTP message (5.3.9). */
typedef struct
{
    UInteger8 lengthField; /**< Length of the text field in the number
    * of octets (5.3.9).
    */
    Octet* textField; /**< Text is encoded as UTF-8 symbols as specified by
    * ISO/IEC 10646:2003 (5.3.9). 
    */
} PTPText;

/** A fault record (5.3.10). */
typedef struct
{
    UInteger16 faultRecordLength; /**< Length of the FaultRecord in number
    * of octets, not including this member (5.3.10).
    */
    Timestamp faultTime; /**< Time at which the fault occurred (5.3.10,
    * 15.5.3.1.7). 
    */
    Enumeration8 severityCode; /**< Severity of the fault (5.3.10, 
    * 15.5.3.1.7).
    */
    PTPText faultName; /**< Name assigned to the fault (5.3.10). */
    PTPText faultValue; /**< Value assigned to the fault (5.3.10). */
    PTPText faultDescription; /**< Description assigned to the fault
    * (5.3.10). 
    */
} FaultRecord;

/** Attributes describing an ordinary clock (8.2.1). */
typedef struct {
    Boolean twoStepFlag; /**< Flag is set to TRUE for two-step clock;
    * FALSE otherwise (8.2.1.2.1). 
    */
    ClockIdentity clockIdentity; /**< Identity of the local clock 
    * (7.6.2.1, 8.2.1.2.2).
    */
    UInteger8 numberPorts; /**< Number of ports on the device, which is
    * always 1 for an ordinary clock (8.2.1.2.3). */
    ClockQuality clockQuality; /**< Quality of the local clock (7.6.2.4,
    * 8.2.1.3.1).
    */
    UInteger8 priority1; /**< Attribute for the priority of the local clock
    * used in the BMC algorithm (7.6.2.2, 8.2.1.4.1). Range of 0 to 255, with 
    * lower values having priority.
    */
    UInteger8 priority2; /**< Attribute for the priority of the local clock
    * used in the BMC algorithm (7.6.2.3, 8.2.1.4.2). Range of 0 to 255, with 
    * lower values having priority. Only used if priority1 values are equal.
    */
    UInteger8 domainNumber; /**< Domain of the local clock (7.1, 
    * 8.2.1.4.3). 
    */
    Boolean slaveOnly; /**< If TRUE, the local clock is a slave-only clock
    * (9.2.2); otherwise, it is a non-slave-only-clock (9.2.3, 8.2.1.4.4).
    */
} DefaultDS;
    
/** Attributes of a clock related to synchronization (8.2.2). */
typedef struct {
    UInteger16 stepsRemoved; /**< Number of communication paths traversed
    * between the local clock and the grandmaster clock (8.2.2.2).
    */
    TimeInterval offsetFromMaster; /**< Best estimate of the current time
    * difference between the master clock and the slave clock as calculated
    * by the slave (8.2.2.3).
    */
    TimeInterval meanPathDelay; /**< Best estimate of the current path
    * delay between the master clock and the slave clock as calculated by 
    * the slave (8.2.2.4).
    */
    TimeInterval lastOffsetFromMaster; /**< Previous offsetFromMaster. */
} CurrentDS;

/** Attributes describing the parent clock (to which the ordinary clock 
 * synchronizes) and the grandmaster clock (at the root of the master-slave
 *hierarchy) (8.2.3).
 */
typedef struct {
    PortIdentity parentPortIdentity; /**< Port identity of the parent 
    * clock (clock to which ordinary clock synchronizes to) (8.2.3.2).
    */
    Boolean parentStats; /**< TRUE if the clock is in SLAVE state and
    * observedParentOffsetScaledLogVariance and observedParentClockPhaseRate
    * are both set to valid estimates, FALSE otherwise (8.2.3.3).
    */
    UInteger16 observedParentOffsetScaledLogVariance; /**< Slave clock's
    * estimate of the variance of the parent clock (8.2.3.4).
    */
    UInteger32 observedParentClockPhaseChangeRate; /**< Slave clock's
    * estimate of the phase change rate of the parent clock with respect
    * to the rate of the slave clock (8.2.3.5).
    * 0x7FFFFFFF if unknown or or too positive for data type, 0x80000000 if
    * too negative for data type.
    */
    ClockIdentity grandmasterIdentity; /**< Identity of the grandmaster
    * clock (8.2.3.6).
    */
    ClockQuality grandmasterClockQuality; /**< Quality of the grandmaster
    * clock (8.2.3.7).
    */
    UInteger8 priority1; /**< Attribute for the priority of the grandmaster
    * clock used in the BMC algorithm (7.6.2.2, 8.2.3.8). Range of 0 to 255, 
    * with lower values having priority.
    */
    UInteger8 priority2; /**< Attribute for the priority of the grandmaster
    * clock used in the BMC algorithm (7.6.2.3, 8.2.3.9). Range of 0 to 255, 
    * with lower values having priority. Only used if priority1 values are
    *equal.
    */
} ParentDS;
 
/** Attributes describing the timescale used for a clock (8.2.4). */
typedef struct {
    UInteger16 currentUtcOffset; /**< Offset between TAI and UTC in units
    * of seconds (only if PTP epoch is used) (8.2.4.2).
    */
    Boolean currentUtcOffsetValid; /**< TRUE if currentUtcOffset is known
    * to be correct, FALSE otherwise (8.2.4.3).
    */
    Boolean leap59; /**< TRUE if the last minute of current UTC day contains
    * 59 seconds (only if PTP epoch is used), FALSE otherwise (8.2.4.4).
    */
    Boolean leap61; /**< TRUE if the last minute of current UTC day contains
    * 61 seconds (only if PTP epoch is used), FALSE otherwise (8.2.4.5).
    */
    Boolean timeTraceable; /**< TRUE if the time is traceable to a primary
    * reference, FALSE otherwise (8.2.4.6).
    */
    Boolean frequencyTraceable; /**< TRUE if the frequency is traceable to a 
    * primary reference, FALSE otherwise (8.2.4.7).
    */
    Boolean ptpTimescale; /**< TRUE if the timescale of the grandmaster
    * clock is PTP, FALSE otherwise (8.2.4.8). 
    */
    UInteger8 timeSource; /**< Source of time (7.6.2.6) used by the 
    * grandmaster clock (8.2.4.9).
    */
} TimePropertiesDS;
    
/** Attributes describing the port of a clock (8.2.5). */
typedef struct {
    PortIdentity portIdentity; /**< Port identity of the local port
    * (8.2.5.2.1).
    */
    Enumeration8 portState; /**< Current state of PTP for the local port
    * (8.2.5.3.1)
    */
    Integer8 logMinDelayReqInterval; /**< Minimum mean time interval
    * (in logarithm base 2) between Delay_Req messages sent by the local
    * port (8.2.5.3.2). This is specified by the master clock based on
    * ability to process Delay_Req messages from all slaves, with an allowed
    * range of [logSyncInterval,logSyncInterval+5].
   */
    TimeInterval peerMeanPathDelay; /**< Current path delay if the delay
    * mechanism is peer-to-peer (currently not implemented), 0 if the delay
    * mechanism is end-to-end (8.2.5.3.3). */
    Integer8 logAnnounceInterval; /**< Mean time interval (in logarithm
    * base 2) between Announce messages. (8.2.5.4.1). This time interval
    * should be equal on ports in a domain.
    */
    UInteger8 announceReceiptTimeout; /**< Number of announce intervals 
    * that must pass without an Announce message being received before an
    * ANNOUNCE_RECEIPT_TIMEOUT_EXPIRE event occurs (8.2.5.4.2). The range is
    * [3,255]. */
    Integer8 logSyncInterval; /**< Mean time interval (in logarithm
    * base 2) between multicast Sync messages sent by the local port
    * (8.2.5.4.3). 
    */
    Enumeration8 delayMechanism; /**< Mechanism used to measure mean
    * path delay (8.2.5.4.4). Only end-to-end is currently implemented.
    */
    UInteger4 versionNumber; /**< Version of PTP being used on the 
    * local port (8.2.5.4.6). Only version 2 is currently implemented.
    */
} PortDS;

/** A management TLV (15.5). */
typedef struct {
    Enumeration16 tlvType; /**< Type is MANAGEMENT (15.5.2.1). */
    UInteger16 lengthField; /**< Length of the data field (must be even)
    * in octets, plus 2. (15.5.2.2).
    */
    Enumeration16 managementId; /**< Management Id (15.5.2.3). */
    Octet* dataField; /**< Data field where the contents depend on the
    * management Id (15.5.3).
    */
    UInteger16 dataFieldLength; /**< Length of the data field in number
    * of octets (15.5.3). 
    */
} ManagementTLV;

/** Fields common for all PTP messages (13.3). */
typedef struct {
    Nibble transportSpecific; /**< Field may be used for the transport protocol
    * and if defined for some transport protocols in annexes D-I (13.3.2.1). 
    */
    Enumeration4 messageType; /**< Type of the message (13.3.2.2). */
    UInteger4 versionPTP; /**< Version of PTP (13.3.2.3). Only version 2
    * is currently implemented. 
    */
    UInteger16 messageLength; /**< Length of the message (13.3.2.4). */
    UInteger8 domainNumber; /**< Domain number, as defined in DefaultDS 
    * (13.3.2.5). 
    */
    Octet flagField[2]; /**< Value of the flag field depends on the
    * message type (13.3.2.6).
    */
    Integer64 correctionField; /**< Value of the correction field depends
    * on the message type, and is expressed in units of nanoseconds 
    * multiplied by 2^16 (13.3.2.7). 
    */
    PortIdentity sourcePortIdentity; /**< Port identity of the port which
    * created the message (13.3.2.8). */
    UInteger16 sequenceId; /**< Value of the sequence Id depends on the
    * message type, but is typically used to identity and order messages
    * (13.3.2.9).
    */
    UInteger8 controlField; /**< Value of the control field depends on the
    * message type (13.3.2.10). */
    Integer8 logMessageInterval; /**< Value of this field depends on the
    * message type (13.3.2.11). */
} Header;

/** Fields for Announce messages (13.5). */
typedef struct {
    Header header; /**< Header of the message (13.3). */
    Timestamp originTimestamp; /**< Time at which the announce message
    * is transmitted (within 1 second accuracy), or 0 (13.5.2.1).
    */
    Integer16 currentUtcOffset; /**< Value of 
    * timePropertiesDS.currentUtcOffset (13.5.2.2).
    */
    UInteger8 grandmasterPriority1; /**< Value of 
    * parentDS.grandmasterPriority1 (13.5.2.3).
    */
    ClockQuality grandmasterClockQuality; /**< Value of 
    * parentDS.grandmasterClockQuality (13.5.2.4).
    */
    UInteger8 grandmasterPriority2; /**< Value of 
    * parentDS.grandmasterPriority2 (13.5.2.5).
    */
    ClockIdentity grandmasterIdentity; /**< Value of 
    * parentDS.grandmasterIdentity (13.5.2.6).
    */
    UInteger16 stepsRemoved; /**< Value of 
    * currentDS.stepsRemoved (13.5.2.7).
    */
    Enumeration8 timeSource; /**< Value of 
    * timePropertiesDS.timeSource (13.5.2.8).
    */
} AnnounceMsg;

/** Fields for Sync messages (13.6). */
typedef struct {
    Header header; /**< Header of the message (13.3). */
    Timestamp originTimestamp; /**< Origin timestamp (9.5.9, 11.3, 
    * 13.6.2.1). 
    */
} SyncMsg;

/** Fields for Follow_Up messages (13.7). */
typedef struct {
    Header header; /**< Header of the message (13.3). */
    Timestamp preciseOriginTimestamp; /**< Origin timestamp (9.5.10, 11.3, 
    * 13.7.2.1).
    */ 
} FollowUpMsg;

/** Fields for Delay_Req messages (13.6). */
typedef struct {
    Header header; /**< Header of the message (13.3). */
    Timestamp originTimestamp; /**< Origin timestamp (9.5.9, 11.3, 
    * 13.6.2.1). 
    */
} DelayReqMsg;

/** Fields for Delay_Resp messages (13.8). */
typedef struct {
    Header header; /**< Header of the message (13.3). */
    Timestamp receiveTimestamp; /**< Reception timestamp of the Delay_Req
    * message which the Delay_Resp message is responding to (9.5.12, 11.3, 
    * 13.8.2.1). 
    */
    PortIdentity requestingPortIdentity; /**< Port identity of the 
    * Delay_Req message which the Delay_Resp message is responding to 
    * (11.3, 13.8.2.2).
    */
} DelayRespMsg;

/** Fields for Management messages (13.13). */
typedef struct { 
    Header header; /**< Header of the message (13.3). */
    PortIdentity targetPortIdentity; /**< Port identity of the port which
    * is the target of the Management message (15.4.1.3).
    */
    UInteger8 startingBoundaryHops; /**< (15.4.1.4). */
    UInteger8 boundaryHops; /**< (15.4.1.5). */
    Enumeration4 actionField; /**< Action to be taken upon receipt of
    * the message (15.4.1.6). 
    */
    ManagementTLV managementTLV; /**< Each management message will have zero
    * or one management TLV (15.4.1.7). */
} ManagementMsg;

/** A common data structure to simplify comparison of clocks by pointing to 
 * values specified in either an Announce message or in the local clock 
 * defaultDS (9.3.4).
 */
typedef struct {
    UInteger8* priority1; /**< Pointing to AnnounceMsg.grandmasterPriority1 or
    * defaultDS.priority1 (9.3.4).
    */
    ClockIdentity *identity;/**< Pointing to AnnounceMsg.grandmasterIdentity or
    * defaultDS.clcokIdentity (9.3.4).
    */ 
    UInteger8* clockClass;/**< Pointing to 
    * AnnounceMsg.grandfatherClockQuality.clockClass or
    * defaultDS.clockQuality.clockClass (9.3.4).
    */
    Enumeration8* accuracy;/**< Pointing to 
    * AnnounceMsg.grandfatherClockQuality.clockAccuracy or
    * defaultDS.clockQuality.clockAccuracy (9.3.4).
    */
    UInteger16* offsetScaledLogVariance; /**< Pointing to 
    * AnnounceMsg.grandfatherClockQuality.offsetScaledLogVariance or
    * defaultDS.clockQuality.offsetScaledLogVariance (9.3.4).
    */
    UInteger8* priority2;/**< Pointing to AnnounceMsg.grandmasterPriority2 or
    * defaultDS.priority2 (9.3.4).
    */
    UInteger16* stepsRemoved;/**< Pointing to AnnounceMsg.stepsRemoved or
    * 0 (9.3.4).
    */
    ClockIdentity* sender;/**< Pointing to AnnounceMsg.sourcePortIdentity or
    * defaultDS.clockIdentity (9.3.4).
    */
    ClockIdentity* receiver;/**< Pointing to 
    * parentDS.portIdentity.clockIdentity or defaultDS.clockIdentity (9.3.4).
    */
    UInteger16* receiverPortNumber;/**< Pointing to 
    * parentDS.portIdentity.portNumber or 0 (9.3.4).
    */
} ClockDataSet;

/** Attributes which describe an event which is used to fire the PTP actor. */
typedef struct {
    Timestamp eventTimestamp; /**< Timestamp at which the event should
    * fire the destination actor.
    */
    UInteger8 eventType; /**< Type of the event. */
    UInteger8 *eventDataPointer; /**< Pointer to event type specific data. */
} Event;

/** Attributes used in qualification of a received Announce message
 *(9.3.2.4).
 */
typedef struct {
    PortIdentity foreignMasterPortIdentity; /**< Port identity of the
    * foreign master (retrieved from Announce message) (9.3.2.4.2).
    */
    UInteger16 foreignMasterAnnounceMessages; /**< Number of Announce
    * messages received from the foreign master within the
    * FOREIGN_MASTER_TIME_WINDOW (4 Announce intervals) time window
    * (9.3.2.4.3).
    */
    AnnounceMsg msg; /**< Last Announce message received from the foreign
    * master.
    */
    Timestamp announceMessageTimestamps[2]; /**< Timestamps of the last two
    * received Announce messages from the foreign master. These timestamps
    * are used to determine whether FOREIGN_MASTER_THRESHOLD (2) Announce 
    * messages have been received within the FOREIGN_MASTER_TIME_WINDOW
    * (4 Announce intervals) time window (9.3.2.4.4).
    */
} ForeignMasterDS;

/** Attributes describing the current state of the PTP. */
typedef struct {

    // PTP data sets. (8.1)
    DefaultDS defaultDS; /**< Describes the ordinary clock (8.2.1). */
    CurrentDS currentDS; /**< Data related to syncrhonization (8.2.2). */
    ParentDS parentDS; /**< Describes the parent and grandmaster clock (8.2.3).     */
    TimePropertiesDS timePropertiesDS; /**< Describes the timescale (8.2.4). */
    PortDS portDS; /**< Describes the port (only 1 port is currently 
    * implemented) (8.2.5).
    */
    
    ForeignMasterDS foreignMasterDS[FOREIGNMASTERDSSIZE]; /**< An array that
    * contains information about foreign masters (9.3.2.4). It is used in
    * the BMC algorithm for qualifying Announce messages.
    */
    UInteger8 foreignMasterDSIndex; /**< Next available index for storing
    * a foreign master in the array. 
    */
    
    UInteger16 sequenceIdAnnounce; /**< Sequence Id for transmitted
    * Announce messages (7.3.7).
    */
    UInteger16 sequenceIdSync; /**< Sequence Id for transmitted
    * Sync messages (7.3.7).
    */
    UInteger16 sequenceIdDelayReq; /**< Sequence Id for transmitted
    * Delay_Req messages (7.3.7).
    */
    UInteger16 sequenceIdManagement; /**< Sequence Id for transmitted
    * Management messages (7.3.7).
    */
    
    Timestamp syncEventIngressTimestamp; /**< Timestamp (according to the
    * slave clock) of the reception of the Sync message.
    */
    Timestamp syncEventEgressTimestamp; /**< Timestamp (according to 
    * master clock) of the transmission of the Sync message.
    */
    Timestamp delayEventIngressTimestamp; /**< Timestamp (according to the
    * master clock) of the reception of the Delay_Req message.
    */
    Timestamp delayEventEgressTimestamp; /**< Timestamp (according to the
    * slave clock) of the transmission of the Delay_Req message.
    */
    TimeInterval syncEventTimeInterval; /**< Time interval between the
    * syncEventEgressTimestamp and the syncEventIngressTimestamp.
    */
    TimeInterval delayEventTimeInterval; /**< Time interval between the
    * delayEventEgressTimestamp and the delaycEventIngressTimestamp.
    */
    Integer64 lastCorrectionField; /**< Last correction field from a
    * message. Correction field is used to correct timestamps.
    */

    // Used in servo algorithm to synchronize slave clock to master clock.
    Integer32 rateAdjFreq; /**< Amount to adjust the rate of the clock
    * to correct the frequency difference between master and slave clocks. 
    * It is expressed in units of 2^-32 nanoseconds per clock cycle.
    */
    Integer32 rateAdjFreqAcc; /**< Accumulator of rateAdjFreq. */
    Integer32 rateAdjOffset; /**< Amount to adjust the rate of the clock
    * to correct the offset between master and slave clocks within specified
    * time period. It is expressed in units of 2^-32 nanoseconds per clock
    * cycle.
    */
    Integer32 rateAdjOffsetAcc; /**< Accumulator of rateAdjOffset. */

    UInteger8 lastAdjType; /**< Last type of clock adjust made (clock step,
    * rate adjust, or combination of both). It is used by adjustClock() to
    * determine they type of the next clock adjustment.
    */
    UInteger16 announceReceiptTimeoutExpiresID; /**< Maintain the Id of the
    * current active announce receipt timeout. Current active timeout
    * can be disabled by incrementing this Id.
    */
    
    // Follow_Up message requires information about last Sync message.
    PortIdentity lastSyncMsgSourcePortIdentity; /**< Source port identity
    * of the last Sync message is needed to determine if a Follow_Up message
    * should be accepted.
    */
    UInteger16 lastSyncMsgSequenceId; /**< Sequence Id of the last Sync 
    * message is needed to determine if a Follow_Up message should be accepted.
    */
    Integer64 lastSyncMsgCorrectionField; /**< Correction field of the last
    * Sync message is used along with the correction field of the accepted
    * Follow_Up message.
    */

    Boolean sameMaster; /**< TRUE if the master clock doesn't change when the
    * BMC algorithm is run, FALSE otherwise. 
    */
    
} PTPState;




