/*
Copyright (c) 2010-2011 The Regents of the University of California. All rights
reserved.

Permission is hereby granted, without written agreement and without license or 
royalty fees, to use, copy, modify, and distribute this software and its 
documentation for any purpose, provided that the above copyright notice and the
following two paragraphs appear in all copies of this software.

IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF
THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF 
CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, 
AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/

/**
 * @file ptp-protocol.h
 * Main operation of IEEE 1588 Precision Time Protocol (PTP).
 *
 * @author Michael Zimmer (mzimmer@eecs.berkeley.edu)
 * @version $Id: ptp-protocol.h 51 2011-04-12 01:19:19Z mzimmer $
 */
 
 #include <stdio.h>
 #include <stdarg.h>
 #include "ptp-constants.h"
 #include "platform/ptp-config.h"
 #include "platform/ptp-platform-data-types.h"
 #include "ptp-data-types.h"
 #include "platform/ptp-msg.h"
 #include "platform/ptp-clock.h"

 

/** The PTPActor() function is responsible for implementing the IEEE 1588 PTP.
 * Correct operation of the protocol is dependent on the PTPActor() being
 * provided with input events when required.
 * Possible input events:
 * - Event provided by powerup of the device:
 *       eventType = EVENT_POWERUP
 * - Event provided by the network protocol stack when a 
 *   new PTP general or event message is received:
 *       eventType = EVENT_MSG_IN
 *       eventDataPointer = \<pointer to message>
 * - Event provided by PTPActor for future firing of itself:
 *       eventType = EVENT_QUALIFICATION_TIMEOUT_EXPIRES |
 *                   EVENT_ANNOUNCE_RECEIPT_TIMEOUT_EXPIRES |
 *                   EVENT_PERIODIC_ANNOUNCE |
 *                   EVENT_PERIODIC_SYNC |
 *                   EVENT_DELAY_REQ |
 *                   EVENT_FOLLOW_UP
 *       eventTimestamp = \<time for event to fire PTPActor>
 * @param[in] eventIn The input event.
 */
void PTPActor(const Event *eventIn);

/** Initialize the data sets in the PTPState \a ptp as specified in 8.2 */
void PTPInitialize();

/** Compute CurrentDS.meanPathDelay. Called when a Sync message (or Follow_Up 
 * message in 2 step mode) is received.
 */
void computeMeanPathDelay();

 /** Compute CurrentDS.offsetFromMaster. Called when a Delay_Resp message
  * is received.
  */
void computeOffsetFromMaster();

/** Adjust slave clock frequency to synchronize to master clock. */
void adjustClock();

/** Create an event to fire PTPActor() at a time 'announce interval' in the
 * future. To cancel this event, 'ptp.announceReceiptTimeoutExpiresID' must be 
 * incremented.
 */
void generateAnnounceReceiptTimeout();

/** Create an event to fire PTPActor() at a time 'N * announce interval' in the
 * future. 
 */
void generateQualificationTimeout(UInteger8 N);

// Functions to handle events and state changes.

/** Handle an incoming PTP message. 
 * @param[in] eventIn The incoming event with a pointer to the message.
 */
void eventMsgIn(const Event *eventIn);

/** Handle an EVENT_POWERUP event (9.2.5, 9.2.6.2). Called when a device is
 * powered on or reset. 
 */
void eventPowerup();

/** Handle an EVENT_INITIALIZE event (9.2.5, 9.2.6.3). Called when an INITIALIZE
 * Management message is received. 
 * */
void eventInitialize();

/** Handle an EVENT_DESIGNATED_ENABLED event (9.2.5, 9.2.6.4). Called when an
 * ENABLE_PORT Management message is received. 
 */
void eventDesignatedEnabled();

/** Handle an EVENT_DESIGNATED_DISABLED event (9.2.5, 9.2.6.5). Called when a
 * DISABLE_PORT Management message is received. 
 */
void eventDesignatedDisabled();

/** Handle an EVENT_FAULT_CLEARED event (9.2.5, 9.2.6.6). Called when fault
 * and/or fualt conditions are cleared. 
 */
void eventFaultCleared();

/** Handle an EVENT_FAULT_DETECTED event (9.2.5, 9.2.6.7). Called when a fault
 * is detected. 
 */
void eventFaultDetected();

/** Handle an EVENT_STATE_DECISION_EVENT event (9.2.5, 9.2.6.8). Called at 
 * least once per 'announce interval' to trigger the BMC algorithm (9.3). 
 */
void eventStateDecisionEvent();

/** Handle an EVENT_QUALIFICATION_TIMEOUT_EXPIRES event (9.2.5, 9.2.6.10). 
 * The EVENT_QUALIFICATION_TIMEOUT_EXPIRES timeout event keeps the protocol in
 *the PRE_MASTER state for a set duration of time.
 */
void eventQualificationTimeoutExpires();

/** Handle an EVENT_ANNOUNCE_RECEIPT_TIMEOUT_EXPIRES event (9.2.5, 9.2.6.11).
 * The EVENT_ANNOUNCE_RECEIPT_TIMEOUT_EXPIRES timout event is used to change
 * the clock state to MASTER if the device hasn't received any Announce messages
 * from other clocks for a set duration of time.
 * @param[in] eventIn The input event.
 */
void eventAnnounceReceiptTimeoutExpires(const Event *eventIn);

/** Handle an EVENT_SYNCHRONIZATION_FAULT event (9.2.5, 9.2.6.12). Called when
 * events in the UNCALIBRATED state should be repeated. 
 */
void eventSynchronizationFault();

/** Handle an EVENT_MASTER_CLOCK_SELECTED event (9.2.5, 9.2.6.13). Called when
 * UNCALIBRATED state has successfully finished.
 */
void eventMasterClockSelected();

/** Handle an EVENT_PERIODIC_ANNOUNCE event. Called periodically with period
 * of 'announce interval'.
 */
void eventPeriodicAnnounce();

/** Handle an EVENT_PERIODIC_SYNC event. Called periodically when clock is
 * MASTER with period of 'sync interval'. */
void eventPeriodicSync();

/** Handle an EVENT_DELAY_REQ event. Called when a Delay_Req message is
 * received.
 */
void eventDelayReq();

/** Handle an EVENT_FOLLOW_UP event. Called when a Follow_Up message needs
 * to be sent (in two step mode).
 */
void eventFollowUp();

/** Perform actions required when entering the INITIALIZING state. */
void stateInitializing();

/** Perform actions required when entering the FAULTY state. */
void stateFaulty();

/** Perform actions required when entering the DISABLED state. */
void stateDisabled();

/** Perform actions required when entering the LISTENING state. */
void stateListening();

/** Perform actions required when entering the PRE_MASTER state. */
void statePreMaster();

/** Perform actions required when entering the MASTER state. */
void stateMaster();

/** Perform actions required when entering the PASSIVE state. */
void statePassive();

/** Perform actions required when entering the UNCALIBRATED state. */
void stateUncalibrated();

/** Perform actions required when entering the SLAVE state. */
void stateSlave();

// Functions to handle received messages.

/** React to a received Announce message (9.5.3). 
 * @param[in] msg The Announce message.
 */
void announceMsgIn(UInteger8* msg);

/** React to a received Sync message (9.5.4). 
 * @param[in] msg The Sync message.
 */
void syncMsgIn(UInteger8* msg);

/** React to a received Follow_Up message (9.5.5).
 * @param[in] msg The Follow_Up message.
 */
void followUpMsgIn(UInteger8* msg);

/** React to a received Delay_Req message (9.5.6). 
 * @param[in] msg The Delay_Req message.
 */
void delayReqMsgIn(UInteger8* msg);

/** React to a received Delay_Resp message (9.5.7). 
 * @param[in] msg The Delay_Resp message.
 */
void delayRespMsgIn(UInteger8* msg);

/** React to a recieved Management message (15). 
 * @param[in] msg The Management message.
 */
void managementMsgIn(UInteger8* msg);

/** Convert message information to common data structure for comparison (9.3.4).
 * @param[in,out] ds The ClockDataSet used in comparison.
 * @param[in] msg The Announce message where the clock information is located.
 */
void toDataSetFromMsg(ClockDataSet *ds, AnnounceMsg *msg);

/** Convert local clock information to common data structure for comparison
 * (9.3.4). 
 * @param[in,out] ds The ClockDataSet used in comparison.
 */
void toDataSetFromLocal(ClockDataSet *ds);

/** Compare two data sets to find which is "better". 
 * @param[in] a A ClockDataSet.
 * @param[in] b A ClockDataSet.
 * @returns The enumeration for which data set is better.
 */
UInteger8 compareDataSet(const ClockDataSet *a, const ClockDataSet *b);

/** Update the data sets according to state decision code (9.3.5).
 * @param[in] code The state decision.
 * @param[in] msg The Announce message containing clock information used to
 * update the \a ptp data sets.
 */
void updateDataSet(UInteger8 code, const AnnounceMsg *msg);

/** Check if two ports are equal. 
 * @param[in] a a A PortIdentity.
 * @param[in] b b A PortIdentity.
 * @returns TRUE or FALSE depending on if PortIdentity equal or not.
 */
Boolean comparePortIdentity(const PortIdentity *a, const PortIdentity *b);

/** Check qualification of Announce message (9.3.2.5).
 * @param d A ForeignMasterDS.
 * @returns TRUE or FALSE if the Announce message in \a d is qualified.
 */
Boolean isQualified(const ForeignMasterDS *d);

/** Fix timestamps when clock is stepped. */
void fixTimestamps(const TimeInterval *timeInterval);

/** Print status to LCD screen. */
void printStatus();
